/*
 *  Copyright (C) 2015  Thadeu Lima de Souza Cascardo <cascardo@minaslivre.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "calcula.h"
#include <errno.h>
#include <stdio.h>
#include "declaracao.h"
#include "cmd.h"
#include "rendimento.h"
#include "totais.h"
#include "util.h"

static const long long dependente2015 = 215652;

long long deducao_dependente(struct declaracao *dec)
{
	if (dec->ano == 2015)
		return dependente2015;
	return 0;
}

/* Alguns totais precisam ser limitados. Portanto, um total de decuções
 * precisa ser ajustado para tais limites. Esta função considerará tais
 * limites no futuro. */
static long long total_deducao(struct declaracao *dec)
{
	int i;
	if (dec->verbose) {
		printf("Dedução:\n");
		printf("\tDependentes: "FMT_R"\n", R(totais_get(dec, "DEPENDENTES")));
		printf("\tINSS: "FMT_R"\n", R(totais_get(dec, "INSS")));
		printf("\tPagamentos: "FMT_R"\n", R(totais_get(dec, "PAGAMENTOS")));
		printf("\tReembolsos: -"FMT_R"\n", R(totais_get(dec, "REEMBOLSOS")));
	}
	return totais_get(dec, "DEPENDENTES") +
	       totais_get(dec, "INSS") +
	       totais_get(dec, "PAGAMENTOS") -
	       totais_get(dec, "REEMBOLSOS");
}

static void total_pago(struct declaracao *dec)
{
	struct rendimento *rendimento;
	int i;
	dec->pago = dec->retido = 0;
	for (i = 0; rendimento = list_get(dec->rendimento, i); i++) {
		dec->pago += rendimento->imposto;
		dec->retido += rendimento->imposto;
	}
	if (dec->verbose) {
		printf("Total pago e retido: "FMT_R" "FMT_R"\n",
			R(dec->pago), R(dec->retido));
	}
}

struct taxtable {
	long long base;
	long long aliquota;
	long long deducao;
};

static struct taxtable table2015[] = {
	{       0,    0,      0, },
	{ 2145324,  750, 160899, },
	{ 3215148, 1500, 402035, },
	{ 4286917, 2250, 723554, },
	{ 5356572, 2750, 991383, },
	{ 9999999999999LL, 0, 0, },
};

static const long long simples2015 = 1588089;

static const long long obrigatoriedade2015 = 2681655;

static long long imposto(struct taxtable *tt, long long tr, int verbose)
{
	int i;
	for (i = 0; tr >= tt[i].base; i++);
	i--;
	if (verbose) {
		printf("Aplicando aliquota de %d%%, deduzindo " FMT_R"\n",
			tt[i].aliquota / 10, R(tt[i].deducao));
	}
	return tr * tt[i].aliquota / 10000 - tt[i].deducao;
}

static long long imposto_simples(struct declaracao *dec)
{
	struct taxtable *tt;
	long long tr, td;
	tt = table2015;
	tr = totais_get(dec, "RENDPJ");
	if (tr / 5 < simples2015)
		td = tr / 5;
	else
		td = simples2015;
	totais_add(dec, "DESCONTO", td);
	tr -= td;
	totais_add(dec, "BASESIMPLES", tr);
	if (dec->verbose) {
		printf("Desconto simplificado é "FMT_R"\n", R(td));
	}
	return imposto(tt, tr, dec->verbose);
}

static long long imposto_completa(struct declaracao *dec)
{
	struct taxtable *tt;
	long long tr, td;
	if (dec->ano != 2015) {
		return -EINVAL;
	}
	tt = table2015;
	tr = totais_get(dec, "RENDPJ");
	td = total_deducao(dec);
	totais_add(dec, "DEDUCOES", td);
	tr -= td;
	totais_add(dec, "BASECOMPLETA", tr);
	if (dec->verbose) {
		printf("Desconto completa é "FMT_R"\n", R(td));
	}
	return imposto(tt, tr, dec->verbose);
}

int calcula(struct declaracao *dec)
{
	long long i_simples, i_completa;
	if (dec->ano != 2015) {
		return -EINVAL;
	}
	if (totais_get(dec, "RENDPJ") > obrigatoriedade2015) {
		if (dec->verbose) {
			printf("Declaracao obrigatoria pois rendimento e"
				"maior que mínimo para declaracao: "
				FMT_R" > "FMT_R"\n",
				R(totais_get(dec, "RENDPJ")),
				R(obrigatoriedade2015));
		}
		dec->obrigatoria = 1;
	}
	i_simples = imposto_simples(dec);
	i_completa = imposto_completa(dec);
	total_pago(dec);
	if (dec->verbose) {
		printf("Imposto simplificada e completa: "FMT_R" "FMT_R"\n",
			R(i_simples), R(i_completa));
	}
	if (dec->tipo != FORCA_SIMPLES &&
	    (i_simples > i_completa || dec->tipo == FORCA_COMPLETA)) {
		totais_add(dec, "BASE", totais_get(dec, "BASECOMPLETA"));
		dec->tipo = COMPLETA;
		dec->devido = i_completa;
	} else {
		totais_add(dec, "BASE", totais_get(dec, "BASESIMPLES"));
		dec->tipo = SIMPLES;
		dec->devido = i_simples;
	}
	if (dec->pago > dec->devido)
		dec->restituicao = dec->pago - dec->devido;
	else
		dec->pagar = dec->devido - dec->pago;
	return 0;
}

static int run_calcula(struct declaracao *dec, char **args, int argc)
{
	totais_add(dec, "EXCLUSIVOS_SEM_13o",
		totais_get(dec, "EXCLUSIVOS") -
		totais_get(dec, "DECIMOTERCEIRO"));
	return calcula(dec);
}

static struct cmd cmd_calcula = {
	.name = "calcula",
	.run = run_calcula,
};

int calcula_cmd_init(void)
{
	cmd_add(&cmd_calcula);
	return 0;
}
